#include "autoap.h"
#include "concretedata.h"
#include "symbian_auto_ptr.h"
#include <charconv.h>
#include <f32file.h>
#include "break.h"
#include "cc_shareddata.h"
#include <cdbcols.h>
#include <commdb.h>

class MIniFileCb {
public:
	/* return ETrue to stop processing */
	virtual TBool SectionBeginL(const TDesC& aSection) = 0;
	virtual TBool KeyL(const TDesC& aName, const TDesC& aValue) = 0;
	virtual TBool UnknownLineL(const TDesC& aLine) = 0;
};

_LIT(KLine, "line");

class CIniFile : public CBase, public MContextBase {
public:
	/* assumes UTF-8 */
	static CIniFile* NewL(MIniFileCb& aCb, const TDesC& aFileName) {
		auto_ptr<CIniFile> ret(new (ELeave) CIniFile(aCb));
		ret->ConstructL(aFileName);
		return ret.release();
	}
	TBool HandleLine(const TDesC& aLine) {
		if (aLine.Length()==0) return EFalse;
		if (aLine[0]==';') return EFalse;
		if (aLine.Length()==1) {
			return iCb.UnknownLineL(aLine);
		}

		if (aLine[0]=='[') {
			if (aLine[aLine.Length()-1]==']') {
				return iCb.SectionBeginL(aLine.Mid(1, aLine.Length()-2));
			} else {
				return iCb.UnknownLineL(aLine);
			}
		}
		TInt equals_pos=aLine.LocateF('=');
		if (equals_pos==KErrNotFound) {
			return iCb.UnknownLineL(aLine);
		} else {
			TInt name_end=equals_pos-1;
			while ( TChar(aLine[name_end]).IsSpace() && name_end>0) 
				name_end --;
			if (name_end==0) {
				return iCb.UnknownLineL(aLine);
			}
			TInt value_begin=equals_pos+1;
			if (value_begin==aLine.Length()) {
				return iCb.KeyL(aLine.Left(name_end+1), KNullDesC);
			}
			while ( TChar(aLine[value_begin]).IsSpace() && value_begin<aLine.Length()-1) 
				value_begin--;
			return iCb.KeyL( aLine.Left(name_end+1), 
				aLine.Right(aLine.Length()-value_begin) );
		}
	}

	void ReadL() {
		TEntry e;
		User::LeaveIfError(Fs().Entry(iFileName, e));
		iPos=0;
		auto_ptr<CCnvCharacterSetConverter> cc(CCnvCharacterSetConverter::NewL());
		CCnvCharacterSetConverter::TAvailability av=
			cc->PrepareToConvertToOrFromL(KCharacterSetIdentifierUtf8, Fs());
		TInt state;
		if (av==CCnvCharacterSetConverter::ENotAvailable) 
			User::Leave(KErrNotSupported);
		do {
			User::LeaveIfError(Fs().ReadFileSection(iFileName,
				iPos, iBuf8, iBuf8.MaxLength()));
			iPos+=iBuf8.Length();
			cc->ConvertToUnicode(iBuf, iBuf8, state);
			for (int i=0; i<iBuf.Length(); i++) {
				if (iBuf[i]=='\n' || iBuf[i]=='\r') {
					iLine->Value().Trim();
					iLine->SyncPtr();
					if (HandleLine( (*iLine)() )) return;
					iLine->Zero();
				} else {
					iLine->Append(iBuf.Mid(i,1));
				}
			}
		} while(iBuf8.Length()>0);
	}
private:
	CBBString* iLine;
	TFileName iFileName;
	TBuf8<128> iBuf8;
	TBuf<256+10> iBuf;
	TInt iPos;
	MIniFileCb& iCb;
	CIniFile(MIniFileCb& aCb) : iCb(aCb) { }
	~CIniFile() {
		delete iLine;
	}
	void ConstructL(const TDesC& aFileName) {
		iLine=CBBString::NewL(KLine);
		iFileName=aFileName;
	}
};

_LIT(KOperaSection, "epoc");
_LIT(KOperaKey, "Connection ID");

class TOperaIniFile : public MIniFileCb {
public:
	TBool iInSection;
	TInt iAp;
	virtual TBool SectionBeginL(const TDesC& aSection) {
#ifdef __WINS__
		RDebug::Print(aSection);
#endif
		if (iInSection) return ETrue;
		if (aSection.CompareF(KOperaSection())==0) iInSection=ETrue;
		return EFalse;
	}
	virtual TBool KeyL(const TDesC& aName, const TDesC& aValue) {
#ifdef __WINS__
		TBuf<101> msg;
		msg=aName.Left(50); msg.Append(_L("=")); msg.Append(aValue.Left(50));
		RDebug::Print(msg);
#endif
		if (aName.CompareF(KOperaKey())==0) {
			TLex l(aValue);
			User::LeaveIfError(l.Val(iAp));
			return ETrue;
		}
		return EFalse;
	}
	virtual TBool UnknownLineL(const TDesC& ) {
		return EFalse;
	}
	TOperaIniFile() : iInSection(0), iAp(-1) { }
};


void GetApNameL(CCommsDatabase& db, TInt aAp, TDes& aName)
{
	aName.Zero(); aName=_L("[not found]");
	CCommsDbTableView* iViewP=db.OpenTableLC(TPtrC(IAP));
	CleanupStack::Pop();
	auto_ptr<CCommsDbTableView> iView(iViewP);
	TUint32 id;
	TInt err;
	while( (err=iView->GotoNextRecord()) == KErrNone ) {
		iView->ReadUintL(TPtrC(COMMDB_ID), id);
		if (id==aAp) {
			iView->ReadTextL(TPtrC(COMMDB_NAME), aName);
			return;
		}
	}
}
void GetApNameL(TInt aAp, TDes& aName)
{
	auto_ptr<CCommsDatabase> db(CCommsDatabase::NewL(EDatabaseTypeIAP));
	GetApNameL(*db, aAp, aName);
}

EXPORT_C TInt CAutoAccessPoint::GetOperaApL(TDes& aName)
{
	_LIT(KOperaIniFile, "\\system\\data\\opera\\opera.ini");

	auto_ptr<HBufC> filename(HBufC::NewL(KOperaIniFile().Length()+2));
	TInt err[2];
	TInt ap=-1;
	filename->Des()=_L("c:");
	for (int i=0; i<2 && ap==-1; i++) {
		TOperaIniFile oi;
		filename->Des().Append(KOperaIniFile);
		auto_ptr<CIniFile> f(CIniFile::NewL(oi, *filename));
		CC_TRAPIGNORE(err[i], KErrNotFound, f->ReadL());
		ap=oi.iAp;
		filename->Des()=_L("e:");
	}
	if (ap!=-1) {
		GetApNameL(ap, aName);
	} else {
		if (err[0]==KErrNone) err[0]=err[1];
		User::LeaveIfError(err[1]);
	}
	return ap;
}

EXPORT_C TInt CAutoAccessPoint::GetBrowserApL(TDes& aName)
{
	aName.Zero();
	RSharedDataClient sh;
	TInt pushed=0;
	CleanupClosePushL(sh); pushed++;
	TInt wapap;
	{
		User::LeaveIfError( sh.Connect() );
		TInt err=sh.Assign( TUid::Uid( 0x10008d39 ) );
		if (err==KErrNotFound) return -1;
		User::LeaveIfError(err);
		err=sh.GetInt(_L("DefaultAP"), wapap);
		if (err==KErrNotFound) return -1;
		User::LeaveIfError(err);
	}
	CleanupStack::PopAndDestroy(pushed);

	{
		auto_ptr<CCommsDatabase> db(CCommsDatabase::NewL(EDatabaseTypeIAP));
		CCommsDbTableView* iViewP=db->OpenTableLC(TPtrC(WAP_IP_BEARER));
		CleanupStack::Pop();
		auto_ptr<CCommsDbTableView> iView(iViewP);
		TInt err;
		while( (err=iView->GotoNextRecord()) == KErrNone ) {
			TUint32 apid;
			iView->ReadUintL(TPtrC(WAP_ACCESS_POINT_ID), apid);
			if (apid==wapap) {
				TUint32 iap;
				iView->ReadUintL(TPtrC(WAP_IAP), iap);
				GetApNameL(*db, iap, aName);
				return iap;
			}
		}
	}
	return -1;
}

_LIT(KInternet, "internet");

EXPORT_C TInt CAutoAccessPoint::GetInternetApL(TDes& aName)
{
	aName.Zero();

	TUint32 service=-1;
	auto_ptr<CCommsDatabase> db(CCommsDatabase::NewL(EDatabaseTypeIAP));
	{
	CCommsDbTableView* iViewP=db->OpenTableLC(TPtrC(OUTGOING_GPRS));
	CleanupStack::Pop();
	auto_ptr<CCommsDbTableView> iView(iViewP);
	TInt err;
	while( (err=iView->GotoNextRecord()) == KErrNone ) {
		iView->ReadUintL(TPtrC(COMMDB_ID), service);
		auto_ptr<HBufC> text(HBufC::NewL(KCommsDbSvrMaxColumnNameLength));
		TPtr ptr(text->Des());
		iView->ReadTextL(TPtrC(GPRS_APN), ptr);
		if ( (*text).CompareF(KInternet)==0) {
			break;
		}
	}
	if (service==-1) return -1;
	}
	{
	CCommsDbTableView* iViewP=db->OpenTableLC(TPtrC(IAP));
	CleanupStack::Pop();
	auto_ptr<CCommsDbTableView> iView(iViewP);
	TUint32 id=-1, this_service;
	TInt err;
	while( (err=iView->GotoNextRecord()) == KErrNone ) {
		iView->ReadUintL(TPtrC(COMMDB_ID), id);
		iView->ReadUintL(TPtrC(IAP_SERVICE), this_service);
		auto_ptr<HBufC> text(HBufC::NewL(KCommsDbSvrMaxColumnNameLength));
		TPtr ptr(text->Des());
		iView->ReadTextL(TPtrC(IAP_SERVICE_TYPE), ptr);
		if ( this_service==service && (*text).CompareF(TPtrC(OUTGOING_GPRS))==0) {
			iView->ReadTextL(TPtrC(COMMDB_NAME), aName);
			return id;
		}
	}
	return -1;
	}
}