#include "pythonexternalizer.h"
#include "errorhandling.h"
#include "symbian_python_ptr.h"

#include <Python.h>
#include <e32base.h>
#include "concretedata.h"

void LeaveIfPythonError(TInt aResult) {
	if (aResult!=0) {
		EnvErr(gettext("Python interpreter returned an error")).
			ErrorCode(MakeErrorCode(CONTEXT_UID_CONTEXTPYTHON, KErrFromPython)).
			Raise();
	}
}
PyObject* LeaveIfPythonNull(PyObject* aObject) {
	if (aObject==0) {
		EnvErr(gettext("Python interpreter couldn't create object")).
			ErrorCode(MakeErrorCode(CONTEXT_UID_CONTEXTPYTHON, KErrFromPython)).
			Raise();
	}
	return aObject;
}

class CPythonExternalizerImpl : public CPythonExternalizer {
public:
	virtual PyObject* GetObject() {
		Py_XINCREF(iTop);
		return iTop;
	}

	struct TItem {
		PyObject*	o;
		TBool		includetype;
		TItem(PyObject* ao, TBool aincludetype) : o(ao), includetype(aincludetype) { }
		TItem(PyObject* ao) : o(ao), includetype(EFalse) { }
		TItem() : o(0), includetype(EFalse) { }
		bool operator != (const TItem& i) const {
			return (o != i.o || includetype != i.includetype );
		}
	};
	CArrayFixSeg<TItem>	*iStack;
	TItem iCurrent;
	PyObject* iTop;
	void ConstructL() {
		iStack=new (ELeave) CArrayFixSeg<TItem>(4);
	}
	virtual void Reset() {
		TItem i;
		while( (i=Pop())!=TItem() ) {
			Py_XDECREF(i.o);
		}
		Py_XDECREF(iTop);
		iCurrent=TItem();
		iTop=0;
	}

	~CPythonExternalizerImpl() {
		Reset();
		delete iStack;
	}
	TItem PushL(TItem o) {
		iStack->AppendL(o);
		return o;
	}
	TItem Pop() {
		if (!iStack || iStack->Count()==0) return TItem();
		TItem o=iStack->At(iStack->Count()-1);
		iStack->Delete(iStack->Count()-1);
		return o;
	}
	void AddToCurrentL(const TDesC& aName, PyObject* aO,
			TBool aIncludeBBType=EFalse) {

		if (aIncludeBBType && !PyDict_Check(iCurrent.o)) {
			Bug(gettext("aIncludeBBType set but no dictionary created")).
				Raise();
		}
		if (iCurrent.o==0) {
			// first
			if (iTop!=0) {
				Bug(gettext("Attempting to serialize to primitives on their own")).
					Raise();
			}
			iTop=aO;
			Py_XINCREF(aO);
		} else if (PyDict_Check(iCurrent.o)) {
			python_ptr<PyObject> name(
				LeaveIfPythonNull( Py_BuildValue("u#", 
					aIncludeBBType ? (TUint16*)L"!value" : (TUint16*)aName.Ptr(), 
					aIncludeBBType ? 6 : aName.Length() )
				));
			LeaveIfPythonError(PyDict_SetItem(iCurrent.o, name.get(), aO));
		} else if (PyList_Check(iCurrent.o)) {
			LeaveIfPythonError(PyList_Append(iCurrent.o, aO));
		} else {
			Bug(gettext("Attempting to add object to something that is neither a list nor dictionary")
				).Raise();
		}
	}
	void BeginType(const TDesC& aName,
			TBool aIncludeBBType, const TTypeName& aBBType) {
		if (!aIncludeBBType) return;
		python_ptr<PyObject> dict( LeaveIfPythonNull(PyDict_New()) );
		python_ptr<PyObject> t( LeaveIfPythonNull(Py_BuildValue("(iiii)",
			aBBType.iModule, aBBType.iId, aBBType.iMajorVersion, aBBType.iMinorVersion)));
		LeaveIfPythonError(
			PyDict_SetItemString(dict.get(), "!bbtype", t.get() ));
		AddToCurrentL(aName, dict.get());
		iCurrent=PushL(TItem(dict.get()));
		dict.release();
	}
	virtual void BeginList(const TDesC& aName,
		TBool aIncludeBBType, const TTypeName& aBBType) {

		BeginType(aName, aIncludeBBType, aBBType);
		python_ptr<PyObject> list( LeaveIfPythonNull(PyList_New(0)) );
		AddToCurrentL(aName, list.get(), aIncludeBBType);
		iCurrent=PushL( TItem(list.get(), aIncludeBBType) );
		list.release();
	}
	virtual void BeginCompound(const TDesC& aName,
			TBool aIncludeBBType, const TTypeName& aBBType) {

		BeginType(aName, aIncludeBBType, aBBType);
		python_ptr<PyObject> dict( LeaveIfPythonNull(PyDict_New()) );
		AddToCurrentL(aName, dict.get(), aIncludeBBType);
		iCurrent=PushL( TItem(dict.get(), aIncludeBBType) );
		dict.release();
	}

	virtual void Field(const TDesC& aName,
			TBasicType aBasicType, const TDesC& aValue,
			TBool aIncludeBBType, const TTypeName& aBBType, 
			TBool aAsAttribute=EFalse) {

		BeginType(aName, aIncludeBBType, aBBType);
		python_ptr<PyObject> val(0);
		_LIT(KVal, "val");
		switch(aBasicType) {
			case EInteger:
				{
					TBBInt i(KVal);
					i.FromStringL(aValue);
					val.reset( LeaveIfPythonNull(PyInt_FromLong(i())) );
				}
				break;
			case EFloat:
				{
					python_ptr<PyObject> str( LeaveIfPythonNull(
						Py_BuildValue("u#", aValue.Ptr(), aValue.Length())) );
					val.reset( LeaveIfPythonNull(PyFloat_FromString(str.get(), 0)) );
				}
				break;
			case EDateTime:
				/* no datetime support in Python 2.2 */
				{
					TBBTime utc(TTime(0), aName);
					utc.FromStringL(aValue);
					utc()-=TTimeIntervalHours(3);
					TBuf<16> s;
					utc.IntoStringL(s);
					val.reset( LeaveIfPythonNull(
						Py_BuildValue("u#", s.Ptr(), s.Length())) );
				}
			case EBinary:
				/* the receiver needs to know that the data is binary... */
			case EString:
				{
					val.reset( LeaveIfPythonNull(
						Py_BuildValue("u#", aValue.Ptr(), aValue.Length())) );
				}
				break;
			case EBoolean:
				{
					PyObject* o;
					TBBBool b(KVal);
					b.FromStringL(aValue);
					if (b()) {
						o=Py_True;
					} else {
						o=Py_False;
					}
					Py_INCREF(o);
					val.reset(o);
				}
				break;
		};
		AddToCurrentL(aName, val.get(), aIncludeBBType);
		if (aIncludeBBType) {
			/* remove the dictionary */
			TItem d=Pop();
			Py_XDECREF(d.o);
		}
	}

	void End() {
		TItem d=Pop();
		Py_XDECREF(d.o);
		if (d.includetype) {
			TItem p=Pop();
			Py_XDECREF(p.o);
		}
	}
	virtual void EndCompound(const TDesC& aName) {
		End();
	}
	virtual void EndList(const TDesC& aName) {			
		End();
	}
	friend CPythonExternalizer;
};

CPythonExternalizer* CPythonExternalizer::NewL()
{
	auto_ptr<CPythonExternalizerImpl> e(new (ELeave) CPythonExternalizerImpl);
	e->ConstructL();
	return e.release();
}
